open Types
open Common
open Print
open String
open List
open Info
open Printf

exception To_string_failed

let debug = Flags.debug_grouper

let heap2s l = List.fold_left (fun s (v,(p,m,_)) -> sprintf "%s%s   >>%d %s<<\n" s (lex2s' v) p (arity2s m)) "" l

let to_string = function IDENT(s,_) | OP(s,_) -> s | _ -> raise To_string_failed
let to_string_ e = try to_string e with To_string_failed -> ""
let extract_pos = function IDENT(_,pos) -> pos | OP(_,pos) -> pos | _ -> failwith "grouper"
let can_mop = function IDENT(_,_) -> true | OP(_,_) -> true | _ -> false
let is_space = function SPACE _ -> true | _ -> false
let l_is_space = function [] -> true | [SPACE _] -> true | _ -> false

let hole = (FQBIGIDENT("Pervasives::Builtin", "Hole", bpos), (9999, `NotOp, `None))    

(********************************************************************************* 
 first pass: regroup simple operators
*********************************************************************************)
let rec l_expr afters befores l = 

  let rec l_expr' l = 
    let (_, heap) = l_expr'' force_reduce' `BadExpr l in
    match map fst heap with
    | [] -> failwith "ooops, don't call l_expr' with empty heap"
    | [ SPACE pos ]
    | _ :: SPACE pos :: _ :: _ ->
	failwith (sprintf "spurious space%s" (pos2sfull pos))

    | [ SPACE _ ; e ; SPACE _ ] | [ SPACE _ ; e ] | [ e ; SPACE _ ] | [e] -> e

    | _ -> failwith ("ooops, reduction failed (" ^ (heap2s heap) ^ ")")


  and l_expr'' force_reduce state l = 
    let (state', heap) = whole_heap (state, []) l in
    let heap' = fold_right may_reduce_from_top heap [] in
    state', force_reduce heap'

  and whole_heap (state, heap) l = 
    if l = [] then (state, heap)
    else 
      let ((state', s_heap), l') = small_heap (state, []) l in
      whole_heap (state', s_heap @ heap) l'

  and start_with_infix e = 
    let e' = match e with
    | LIST([], e'::_) -> e'
    | _ -> e
    in let (_, kind, _) = what_kind e' `Expr
    in kind = `Infix

  and small_heap (state, heap) ll = 
(*    if debug then printf "small_heap %s [[[[[[%s]]]]]]\n" (llex2s ll) (heap2s heap);*)
    match ll with
  | [] -> (state, force_reduce (fold_right may_reduce_from_top heap [])), []
  | SPACE pos :: l -> (state, (SPACE pos, (9999, `NotOp, `None)) :: force_reduce heap), l
  | OP("magic;", _) :: l when state = `BadExpr -> (state, force_reduce heap), l
  | OP("magic;", _) :: e :: l when start_with_infix e -> (state, force_reduce heap), e::l
  | OP("magic;", _) :: l -> small_heap (state, heap) (if l = [] then [] else OP(";;", bpos) :: l)
  | LIST([], l1) :: l -> 
      let (state', heap') = l_expr'' force_reduce state l1 
      in small_heap (state', heap' @ heap) l

  | LIST(["(",pos1 ; ")",pos2] as tl, l') :: l -> 
      if l_is_space l' then
	small_heap (state, heap) (FQBIGIDENT("Pervasives::Builtin", "Empty_tuple", unipos pos1 pos2) :: l)
      else
        (* keeping those parentheses for better structuring. Cleaned away in function multi_ops *)
	let e = LIST(tl, [ l_expr' l' ]) in
	small_heap (if l' = [] then (state, heap) else add2heap e state heap) l
  | LIST(name, l') :: l -> 
      let op = 
	let prefix = if state =`Expr then "multiple_suffix_" else "multiple_" in
	let join = if l_is_space l' then "" else "_" in
	OP(prefix ^ String.concat join (map fst name), uniposl (map snd name)) in
      let state', heap' =
	if state = `Expr then
	  (* special case (cough!), for around-ops acting as suffix... *)
	  let state2, heap2 = add2heap (OP(".", bpos)) state heap in
	  let state3, heap3 = add2heap op state2 heap2 in
	  if l_is_space l' then state3, heap3 else add2heap (LIST(["(",bpos ; ")",bpos], [ l_expr' l' ])) state3 heap3
	else
	  let e = if l_is_space l' then op else CALL(op, l_expr' l') in
	  add2heap e state heap
      in
      small_heap (state', heap') l

  | e :: l -> 
      small_heap (add2heap e state heap) l

  and force_reduce heap = 
(*    if debug then printf "forcing reduction of heap %s**************************************************************\n" (heap2s heap);*)
    may_reduce (-9998) `None heap

  and force_reduce' heap = 
(*    if debug then printf "forcing reduction of heap %s**************************************************************\n" (heap2s heap);*)
    may_reduce (-9999) `None heap
	  
  and add2heap e state heap =
    let (prio, kind, asso) as t = what_kind e state in
    let (state, heap) =
      if state = `BadExpr && mem kind [ `Separator ; `Infix ; `Suffix]
      then (`Expr, hole :: heap)
      else (state, heap) in

    if debug then Printf.printf "ADDing %s (%s, %s, %d)\n" (if state = `Expr then "Expr" else "BadExpr") (lex2s e) (arity2s kind) prio;

    let state' = if mem kind [ `Infix ; `Separator ; `Prefix ] then `BadExpr else `Expr in
    state', (e, t) :: may_reduce prio asso heap

  and what_kind e state = 
    try assoc (to_string e) (match state with
    | `Expr -> afters @ befores
    | `BadExpr -> befores @ afters)
    with Not_found | To_string_failed -> (9999, `NotOp, `None)

  and is_also_suffix e = exists (fun (_,where,_) -> where = `Suffix) (all_assoc (to_string e) afters)

  and may_reduce_from_top e l =
    if debug then print_string "may_reduce_from_top\n"; 
      let (_, (prio, kind, asso)) = e in 
      e :: if kind = `Prefix then l else may_reduce prio asso l

  and may_reduce prio asso heap =
    if debug then (
      printf "-------------may_reduce (%d)---------------------------------------------------------\n" prio;
      print_string (heap2s heap);
      print_string "\n"
     );
    
    let verif prio2 e = prio2 > prio || prio2 = prio && match asso with
    | `None -> failwith (sprintf "non associative operator ``%s''%s" (lex2s e) (pos2sfull (extract_pos e)))
    | `Right -> false
    | `Left -> true

    in let reduce para e l = may_reduce prio asso ((CALL(e, para), (9999, `NotOp, `None)) :: l)

    in let add2sep sep a b =
      let s2l = function
	| SEP(sep', l) when sep = sep' -> l
	| e -> [e]
      in
      SEP(sep, s2l a @ s2l b)
	
    in 
    (match heap with
      (* remove SPACE bubbles *)
    | (_, (_, kind, _) as op) :: (SPACE _, _) :: l when kind <> `NotOp -> 
	may_reduce prio asso (op :: l)
    | e :: (_, (_, kind, _) as op) :: (SPACE _, _) :: l when kind <> `NotOp -> 
	may_reduce prio asso (e :: op :: l)
    | (SPACE _, _) :: (opname, (_, kind, _) as op) :: l when kind <> `NotOp ->
	(match opname with
	| OP(".", pos) -> failwith (sprintf "no space allowed between the ``.'' and the method name%s" (pos2sfull pos))
	| _ -> may_reduce prio asso (op :: l))

    | (e, (_, `Suffix, _)) :: (e2, _) :: l (* don't care priority, we're doing it, para is reduced already *)
      -> reduce e2 e l
    | (e2, _) :: (e, (prio2, `Prefix, _)) :: l when verif prio2 e
      -> reduce e2 e l
    | (e2, (_, kind, _)) :: (e, (prio2, `Infix, _)) :: (e3, _) :: l when verif prio2 e && kind <> `Prefix
      -> reduce (SEP(",", [e3 ; e2])) e l
    | (e2, (_, kind, _)) :: (e, (prio2, `Separator, _)) :: (e3, _) :: l when verif prio2 e && kind <> `Prefix
      -> may_reduce prio asso ((add2sep (to_string e) e3 e2, (9999, `NotOp, `None)) :: l)

    | (e1, (_, `NotOp, _)) :: (e2, (_prio2, `NotOp, _)) :: l when not (is_space e1 || is_space e2)
      -> reduce e1 e2 l

    (* special case for infix AND suffix operator: rewrite as suffix *)
    | (e, (prio2, `Infix, asso)) :: l when prio = (-9999) && is_also_suffix e -> 
	may_reduce prio asso ((OP ("suffix" ^ to_string e, extract_pos e), (prio2, `Suffix, asso)) :: l) 

    (* add a hole in the heap *)
    | (_sep, (_, t, _)) :: _ when prio = (-9999) && mem t [`Separator ; `Prefix ; `Infix ] -> 
	may_reduce prio asso (hole :: heap) 
    | [ _, (_, `Suffix, _) ] ->
	may_reduce prio asso (heap @ [hole])

    | e::l -> e :: may_reduce prio asso l
    | [] -> [])
      
  in l_expr' l

(********************************************************************************* 
 second pass: regroup multiple operators
*********************************************************************************)
let rmm multi_ops e =
  let multi_ops' = map fst (sort (fun (_,l1) (_,l2) -> if l1 < l2 then 1 else -1) (map (fun e -> e, length e) multi_ops)) in

  let rec search_m_op = function
    | [], e -> [], [e]
    | ops, (CALL(op, SEP(",", [ a ; b ])) as e) ->
	let bo = search_m_op(ops, a) in
	(match bo with
	| op' :: ops', collected when op' = to_string_ op ->		  
	    let ops'', collected2 = search_m_op(ops', b) in
	    ops'', collected @ collected2
	| _ -> ops, [e])
    | op'::ops, CALL(op, b) when op' = to_string_ op ->
	search_m_op(ops, b)
    | ops, e -> ops, [e]
  in
  let rec group e = match e with

  | CALL(f, para) ->
      (try
	let ops, paras = find_some (fun ops -> 
	  let ops', paras = search_m_op(ops, e) in
	  if ops' = [] then Some(ops, paras) else None
	) multi_ops' in
	
	if debug then printf "found (( %s ))\n" (String.concat "|" ops);
	CALL(IDENT(String.concat "_" ("multiple" :: ops), Parser.getpos para), SEP(",", map group paras))
      with Not_found ->
	if debug then printf "not found %s\n" (lex2s f);
	CALL(group f, group para))

  | SEP(sep, l) -> SEP(sep, map group l)	
  | LIST(name, l) -> LIST(name, map group l)

  | SPACE _
  | EOF
  | INDENT _
  | INDENT_ _
  | _ -> e
 
  in group e


(********************************************************************************* 
  cleanup, move to new format
*********************************************************************************)
let rec cleanup e = match e with
  | CALL (OP(".",_), SEP (",", [e ; CALL (metho, LIST(["(",_ ; ")",_], [ e' ]))])) -> 
      let para = (match e' with
      |	SEP(",", l) -> l
      |	e -> [e]) in
      cleanup (CALL(metho, SEP(",", e :: para)))

  | CALL (OP(".",_), SEP (",", [e ; e'])) -> cleanup (CALL(e', e))

  | CALL(e, e') -> `Call(cleanup e, cleanup e')

  (* those parentheses were kept during grouping for better structuring *)
  (* double parentheses mean we really want a singleton Sep *)
  | LIST(["(",_ ; ")",_], [ LIST(["(",_ ; ")",_], [e]) ]) -> `Sep(",", [ cleanup e ])
  | LIST(["(",_ ; ")",_], [e]) -> cleanup e

  | OP(e,pos) -> `Fqid("", e, pos)
  | IDENT(e,pos) -> `Id(e, pos)
  | BIGIDENT(e,pos) -> `Constr(e, pos)
  | NUM(e,pos) -> `Float(e, pos)
  | STRING(e,pos) -> `String(e, pos)
  | FQIDENT(fq, e,pos) -> `Fqid(fq, e, pos)
  | FQBIGIDENT(fq, e,pos) -> `Fqconstr(fq, e, pos)

  | SEP(";;", l) -> cleanup (SEP(";", l))
  | SEP(sep, l) -> `Sep(sep, map cleanup l)
  | SPACE _
  | EOF
  | INDENT _
  | INDENT_ _
  | LIST _ -> failwith (sprintf "go away, you don't exist (%s)" (lex2s e))



(********************************************************************************* 
 do the 3 passes
*********************************************************************************)
let (grouper : all_operators -> fromlexer list -> fromgrouper) =
  fun ops l ->
    match l with
    | [] -> `Fqconstr("Pervasives::Builtin", "Empty_tuple", bpos)
    | _ ->
	let e = l_expr ops.afters_op ops.befores_op l in
	let e' = rmm ops.multi_ops_op e in 
	if debug then printf "grouper: %s\n" (lex2s e');
	cleanup e'
