open Types
open Common
open Printf
open Print
open String
open Info
open List

let debug = Flags.debug_parser
let tabsize_default = 8
let pervasives_imports = list_to_StringSet [ "Pervasives" ]

let is_space = function SPACE _ -> true | _ -> false

let rec getpos = function
  | SPACE(pos)
  | OP(_,pos)
  | IDENT(_,pos)
  | FQIDENT(_,_,pos)
  | BIGIDENT(_,pos)
  | FQBIGIDENT(_,_,pos)
  | NUM(_,pos)
  | STRING(_,pos) -> pos
  | CALL(e1, e2) -> unipos (getpos e1) (getpos e2)
  | LIST(l1, l2) -> uniposl (map snd l1 @ map getpos l2)
  | SEP(_, l) -> uniposl (map getpos l)
  | EOF
  | INDENT_ _
  | INDENT _ -> bpos

(*********************************************************************************
* first pass: group parentheses 
*********************************************************************************)
let parse_arounds arounds l =
  let arounds' = map swap arounds in

  let rec parse = function
    | OP(a, pos) :: l ->
	(try
	  let b = assoc a arounds in
	  let (list, l', last) = parse l in
	  match last with
	  | None -> 
	      failwith (sprintf "missing ending for around op ``%s''%s" a (pos2sfull pos))
	  | Some (b', pos') -> if b = b' then
	      next (LIST([a,pos ; b,pos'], list)) l'
	      else failwith (sprintf "``%s'' does not correspond to ``%s''" b' a)
	with Not_found -> 
	  try
	    let _ = assoc a arounds' in [], l, Some (a, pos)
	  with Not_found -> next (OP(a, pos)) l)
    | LIST(name, l1) :: l -> next (LIST(name, parse' l1)) l
    | other :: l -> next other l
    | [] -> [], [], None

  and next e l =
    let (l', rest, last) = parse l in 
    e::l', rest, last

  and parse' l =
    if debug then printf "parsing (%s)\n" (String.concat "," (map lex2s l));
    match parse l with
    | (l, [], None) -> l
    | (_, _, Some (last, pos)) -> failwith (sprintf "Ending ``%s'' with no opening%s" last (pos2sfull pos))
    | _ -> failwith "parse'"

  in parse' l

(* call it with ``parse_arounds ["(", ")"] l'' *)


(********************************************************************************* 
* second pass: collect and remove pragmas
*              assign INDENT's value
*              form full qualified idents and constructors
*********************************************************************************)

let parse_pragmas l =
  let arities =
    [ 
      "around",  (`Around,  `None) ;
      "suffix",  (`Suffix, `Right) ;
      "prefix",  (`Prefix, `Right) ;
      "separator", (`Separator, `Right) ;
      "infix",   (`Infix,  `None) ;
      "infixr",  (`Infix,  `Right);
      "infixl",  (`Infix,  `Left) ;
    ]
  in
  let to_string = function IDENT(s,_) -> s | OP(s,_) -> s | _ -> failwith "to_string"
  in
  let indent_size s tabsize =
    let l = ref 0 in
    for i = 0 to String.length s - 1 do
      match get s i with
      | ' '  -> incr l    
      | '\t' -> l := adjustModUp tabsize (!l + 1)
      | _ -> internal_error "indent_size"
    done ; !l
  in
  let rec parse_pragmas' (ops, arounds, tabsize, imports, packages_used as alls) = 
    let parse_pragma pos para = function
      | "tabsize" ->
	  (match para with
	  | [ NUM(n,_) ] when is_int n -> 
	      if tabsize = None then ops, arounds, Some (int_of_float n), imports, packages_used
	      else failwith ("you can't set Pragma::tabsize twice in same file" ^ pos2s pos)
	  | _ -> failwith ("Pragma::tabsize wants a number as argument" ^ pos2s pos)
	  )
      |	"import" ->
	  let rec collect_bigidents = function
	    | [ BIGIDENT("Vars", _) ] -> []
	    | BIGIDENT(package, _) :: OP("::", _) :: l -> package :: collect_bigidents l
	    | _ -> failwith ("usage of Pragma::import is Pragma::import(Package_name::Vars)" ^ pos2sfull pos)
	  in
	  let package = String.concat "::" (collect_bigidents para) in
	  ops, arounds, tabsize, stringSet_add imports package, stringSet_add packages_used package

      |	"no_pervasives_import" -> 
	  ops, arounds, tabsize, stringSet_difference imports pervasives_imports, stringSet_difference packages_used pervasives_imports

      |	"around" ->
	  (match para with
	  | [ beg ; end_ ] -> ops, (to_string beg, to_string end_) :: arounds, tabsize, imports, packages_used
	  | _ -> failwith ("Pragma::around must be called Pragma::around(<open op> <close op>)" ^ pos2s pos))
      | e ->
	  try
	    let (arity, asso) = assoc e arities in
	    (match para with 
	    | OP("-",_) :: NUM(prio,_) :: OP(",",_) :: (_ :: _ as op) when is_int prio ->
		(map to_string op, (- int_of_float prio, arity, asso)) :: ops, arounds, tabsize, imports, packages_used
	    | NUM(prio,_) :: OP(",",_) :: (_ :: _ as op) when is_int prio ->
		(map to_string op, (int_of_float prio, arity, asso)) :: ops, arounds, tabsize, imports, packages_used
	    | _ -> failwith (Printf.sprintf "Pragma::%s must be called Pragma::%s(<priority>,<operator>) with no fancy%s\n(especially, don't put spaces along the comma, and <priority> must be a number)" e e (pos2s pos)))
	  with Not_found -> failwith ("unknown pragma Pragma::" ^ e ^ pos2s pos)
	      
    in function
      | [] -> alls, []
      | BIGIDENT("Pragma", pos1) :: OP("::", _) :: IDENT(name, _) :: LIST(["(",_ ; ")",_], para) :: l ->
	  let pos = uniposl (pos1 :: (map getpos para)) in
	  parse_pragmas' (parse_pragma pos (filter (o not is_space) para) name)  l
      |	BIGIDENT(q,pos) :: OP("::", _) :: l ->
	  let rec fqualify q pos = function
	    | BIGIDENT(q',pos') :: OP("::", _) :: l -> fqualify (q ^ "::" ^ q') (unipos pos pos') l
	    | l -> q, pos, l in
	  let q, pos, l = fqualify q pos l in

	  let fully_qualified, l = 
	    match l with
	    | IDENT   (i, pos') :: l
	    | OP      (i, pos') :: l
	    | STRING  (i, pos') :: l -> FQIDENT   (q, i, unipos pos pos'), l
	    | BIGIDENT(i, pos') :: l -> FQBIGIDENT(q, i, unipos pos pos'), l
	    | e -> failwith (sprintf "``<module>::'' must be followed by an ident (found %s)" (llex2s e))
	  in
	  let packages_used = if q = "" then packages_used else stringSet_add packages_used q in
	  parse_pragmas' (ops, arounds, tabsize, imports, packages_used) (fully_qualified :: l)

       (* remove double INDENT (happens with comments) *)
      |	INDENT _ :: (INDENT _ :: _  as l) ->
	  parse_pragmas' alls l

      |	OP("::", pos) :: _ as l -> 
	  (* insert an "" BIGIDENT *)
	  parse_pragmas' alls (BIGIDENT("", pos) :: l)

      | e :: l -> 
	  let (_, _, tabsize, _, _ as alls), l = parse_pragmas' alls l in 

	  (match e with
	  | (LIST(around, l2)) -> 
	      let alls, l2 = parse_pragmas' alls l2 in 
	      alls, LIST(around, l2) :: l
	  | INDENT(s) -> 
	      alls, INDENT_(indent_size s (some_or tabsize tabsize_default)) :: l
	  | _ ->
	      alls, e :: l)
  in

  let add_op l (e,v) =
    try
      let v' = assoc e l in
      if v = v' 
      then l
      else match (v, v') with
      |	(_, `Infix, _), (_, `Suffix, _) -> (e,v) :: l
      |	_ -> failwith (sprintf "operator %s is already defined (you can redefine an operator if same priority (!! suffix must be before infix !!))" e)
    with Not_found -> (e,v) :: l
  in
  let (collect_small_ops : string list * (int * where * asso) -> 'a * 'a -> 'a * 'a) = 
    fun (op, (prio, arity, asso)) (afters, befores) ->
      let mlen = (length op) - 1 in
      match arity with
      | `Separator -> 
	  if length op > 1 then failwith "Pragma::separator must be called Pragma::separator(<priority>,<operator>)"
	  else add_op afters (hd op, (prio, arity, asso)), add_op befores (hd op, (prio, arity, asso))
      | `Prefix -> 
	  fold_left add_op afters (map (fun e -> e, (prio, `Infix, asso)) (tl op)),
	  add_op befores (hd op, (prio, `Prefix, asso))
      | `Infix -> fold_left add_op afters (map2 (fun e asso -> e, (prio, `Infix, asso))
					     op (asso :: repeat `Right mlen)), befores
      | `Suffix -> fold_left add_op afters (map2 (fun e ari -> e, (prio, ari, asso)) op (repeat `Infix mlen @ [`Suffix])), befores
      | _ -> failwith "unknown arity"
  in
  let (ops, arounds, _tabsize, imports, packages_used), l' = parse_pragmas' ([], [], None, pervasives_imports, pervasives_imports) l in
  let (afters, befores) = fold_right collect_small_ops ops ([],[]) in
  { arounds_op = arounds ; afters_op = afters ; befores_op = befores ; multi_ops_op = filter (fun l -> length l > 1) (map fst ops) }, 
   rev (stringSet_to_list imports), rev (stringSet_to_list packages_used), l'

(********************************************************************************* 
* third pass: group using around ops (parentheses already done at first pass)
*********************************************************************************)
(* reuse function for parentheses, aka call ``parse_arounds arounds l'' *)

(********************************************************************************* 
* fourth pass: take care of indentation
*********************************************************************************)
let indent afters befores l = 
  let afters' =  difference (map fst afters) (map fst befores) in
  let befores' = map fst (befores @ filter (fun (_, (_, arity, _)) -> arity = `Infix || arity = `Separator) afters) in

  let mymem l = function
  | IDENT(e,_) -> mem e l
  | OP(e,_) -> mem e l
  | _ -> false
  in

  let make_list = function
    | [] -> []
    | [ LIST _ ] as e -> e
    | e -> [ LIST([], e) ]
  in 

  let is_separator = function
    | OP(e,_) -> 
	(try 
	  let _, kind, _ = assoc e befores in
	  kind = `Separator
	with Not_found -> false)
    | _ -> false
  in

  let rec parse = function
    | [] -> [], []
    | [(_, _, l)] -> l, []
    | (old_s, old_last, old_l) :: (s, last, l) :: ll ->
	if debug && (l <> [] || old_l <> []) then (
	  Printf.printf "sizes, before : %d   after %d\n" old_s s;
	  print_string "prev line: "; print_string (llex2s old_l);
	  print_string "next line: "; print_string (llex2s l ^ "\n");
	);
	if s = old_s then
	  let (v, ll') = parse ((s,last,l) :: ll) in
	  if mymem befores' last || l <> [] && mymem afters' (hd l)
	  then LIST([], old_l) :: v, ll'
	  else if old_l = [] 
	       then v, ll'
	       else if v = []
                    then old_l, ll'
                    else LIST([], old_l) :: SPACE(bpos) :: OP("magic;", bpos) :: SPACE(bpos) :: make_list(v), ll'
	else if s > old_s then
	  let (v, ll') = parse ((s,last,l) :: ll) in
	  let magic_or_not = 
	    if mymem befores' last || l <> [] && mymem befores' (hd l) 
	    then [] 
	    else [SPACE(bpos) ; OP("magic;", bpos) ; SPACE(bpos) ] in
	  if v = [] then
	    parse ((old_s, old_last, old_l) :: ll')
	  else 
	    parse ((old_s, old_last, old_l @ magic_or_not @ [ LIST([], v) ]) :: ll')
	else 
	  let old_l' = if is_separator last then removelast old_l else old_l in
	  let last'  = if is_separator last then Common.last old_l' else last in
	  [ LIST([], old_l') ], (s, last', l) :: ll

  and lines last = function
    | [] -> ([], [])
    | e :: l -> 	
	let (l1, l2) = lines e l in
	match e with
	| INDENT_ s -> [], (s, last, l1) :: l2
	| LIST(name, l) -> LIST(name, parse' (lines' l)) :: l1, l2
	| _ -> e :: l1, l2

  and parse' l = fst (parse l)
  and lines' l = 
    let (beg, l') = lines (SPACE bpos) l in 
    if beg = [] then l' else (0, SPACE bpos, beg) :: l'

  in let l' = (lines' l)
  in parse' l'

let add_ops ops1 ops2 =
  { 
    arounds_op   = ops1.arounds_op   @ ops2.arounds_op ;
    befores_op   = ops1.befores_op   @ ops2.befores_op ;
    afters_op    = ops1.afters_op    @ ops2.afters_op ;
    multi_ops_op = ops1.multi_ops_op @ ops2.multi_ops_op ;
  } 

let ops_empty = { arounds_op = [] ; befores_op = [] ; afters_op = [] ; multi_ops_op = [] }

let add_pretty_printer_ops ops =
  Print.operators := 
    filter_some (function
      |	n, (prio, kind, _) when kind = `Infix || kind = `Separator -> Some(n, prio)
      | _ -> None
    ) ops.afters_op

(********************************************************************************* 
* do the 4 passes
*********************************************************************************)
let parse_all1 l =
  let l2 = parse_arounds ["(", ")"] l in
  parse_pragmas l2

let parse_all2 ops l =
  let l2 = parse_arounds ops.arounds_op l in
  let l3 = indent ops.afters_op ops.befores_op l2 in
  if debug then (printf "parsed:\n%s\n" (String.concat "\n" (map lex2s l3)));
  add_pretty_printer_ops ops ;
  l3
