open Types
open Printf
open Common
open List
open String

let verbose = Flags.verbose_print

let operators : (string * int) list ref = ref []

let list2s sep l = "(" ^ concat sep l ^ ")"

let list2s_ sep = function
  | [e] -> e
  | l -> list2s sep l

let pos2s pos = 
  try
    let (file, line, _, _) = Info.pos2line pos in
    sprintf " (%s:%d)" file line
  with Not_found -> ""

let rec (lex2s : fromlexer -> string) = function
   | OP(";", _) -> ";\n"
   | OP(s, _) -> s
   | SPACE(_) -> " "
   | INDENT s -> s
   | IDENT(s,_) -> s
   | FQIDENT(q, s, _) -> q ^ "::" ^ s
   | BIGIDENT(s, _) -> s
   | FQBIGIDENT(q, s, _) -> q ^ "::" ^ s
   | NUM(f, _) -> string_of_float f
   | STRING(s, _) -> "\"" ^ s ^ "\""
   | EOF -> "EOF"

   | LIST([a,_;b,_], l) -> a ^ (llex2s l) ^ b
   | LIST(name, l) -> sprintf "list%s(%s)" (concat "" (map fst name)) (llex2s l)
   | SEP(name, l) -> "(" ^ concat name (map lex2s l) ^ ")"

   | CALL(name, para) -> sprintf "%s(%s)" (lex2s name) (lex2s para)
   | INDENT_ _ -> ""

and (lex2s' : fromlexer -> string) = function
   | OP(s, pos) -> s ^ (pos2s pos)
   | IDENT(s,pos) -> s ^ (pos2s pos)
   | FQIDENT(q, s, pos) -> q ^ "::" ^ s ^ (pos2s pos)
   | BIGIDENT(s, pos) -> s ^ (pos2s pos)
   | FQBIGIDENT(q, s, pos) -> q ^ "::" ^ s ^ (pos2s pos)
   | NUM(f, pos) -> string_of_float f ^ (pos2s pos)
   | STRING(s, pos) -> "\"" ^ s ^ "\"" ^ (pos2s pos)
   | e -> lex2s e

and llex2s l = concat " " (List.map lex2s l) 
and llex2s' l = concat " " (List.map lex2s l) 


let (arity2s : where -> string) = function
  | `Infix  -> "Infix" 
  | `NotOp  -> "NotOp" 
  | `Suffix -> "Suffix"
  | `Around -> "Around"
  | `Prefix -> "Prefix"
  | `Separator -> "Separator"



let e2s =  
  let rec (e2s' : fromgrouper -> string) = 
    function
      | `Constr(e,_) -> e
      | `Call(e,e') -> sprintf "%s(%s)" (e2s' e) (e2s' e')
      | `Float(f,_) -> string_of_float f
      | `String(s,_) -> sprintf "\"%s\"" s
      | `Sep(e,l) -> concat e (map e2s' l)
      | `Id(e,_) -> e
      | `Fqid(l,e,_) -> l ^ "::" ^ e
      | `Fqconstr(l,e,_) -> l ^ "::" ^ e
  in function
    | `Sep(e,l) -> concat (e ^ "\n") (map e2s' l)
    | e -> e2s' e

let pr f e = print_string (f ^ ": " ^ (e2s e) ^ "\n")


let io2s = function
  | In -> "In"
  | Out -> "Out"
  | Exact -> "Exact"
  | Inout -> "Inout"

let rew2s_ verbose deref = 
  let is_an_operator = function
    | Global((_, _, s),_) ->
	(try let _ = assoc s !operators in Some s
	with Not_found -> None)
    | _ -> None
  in
  
  let rec handle_op op l =
    let prio = 
      try assoc op !operators
      with Not_found -> internal_error (sprintf "bad operator %s" op) in
    let l' = map (fun (s, prio') -> if prio' <= prio then "("^s^")" else s) l in
    if op = "," then
      String.concat (op ^ " ") l', prio
    else
      let compact_is_nice = fold_left min 999 (map snd l) in
      if compact_is_nice = 999 then
	String.concat op l', 998
      else
	String.concat (" " ^ op ^ " ") l', prio

  and r_ e = fst (r e)

  and r = function
    | Constr("_", e, _) -> e, 999
    | Constr("Pervasives::Builtin", "Empty_tuple", _) -> "()", 999
    | Constr(q ,s,_) -> q ^ "::" ^ s, 998
    | Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) -> "_", 998
    | Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], l)) -> 
	let l', var = split_last l in
	sprintf "(%s where %s)" (r_ var) (String.concat " and " (map r_ l')), 998
    | Call(Constr("Pervasives::Builtin", "Var_exists", _), SepL(";", [_], l)) -> 
	let l', var = split_last l in
	sprintf "(one %s where %s)" (r_ var) (String.concat " and " (map r_ l')), 998
    | Call(Constr("Pervasives::Builtin","Abstract",_), Constr(_, a, _)) -> "&" ^ a, 998
    | Call(Constr("Pervasives::Builtin","Or_range",_), Sep(",", l)) -> concat "...." (map r_ l), 998

    | Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", [])) -> "Object", 999
    | Call(Constr("Pervasives::Builtin", "Or",    _), Sep(",", [])) -> "Subval", 999

    | Call(Constr("Pervasives::Builtin","All_T",_), Sep(",", l)) -> handle_op "|&|" (map r l)
    | Call(Constr("Pervasives::Builtin","Or",_), Sep(",", l)) -> handle_op "|" (map r l)
    | Call(Call(Constr("Pervasives::Builtin", "Val", _), Global((_, _, "!>"),_)), Sep(",", [a;b])) -> handle_op "!>" (map r [a;b])
    | Call(Call(Constr("Pervasives::Builtin", "Val", _), Global((_, _, "!<"),_)), Sep(",", [a;b])) -> handle_op "!<" (map r [a;b])
    | Call(Constr("Pervasives::Builtin", "Overload", _), Sep(",", l)) -> 
	"Overload" ^
	concat "\n+" (map (function 
	  | Sep(",", [ value ; typ ]) -> "\n  " ^ r_ value ^ "\n  " ^ r_ typ 
	  | _ -> failwith "unoverload"
	) l), 998
    | Call(e, e') -> 
	(match is_an_operator e with
	| None -> 
	    let e_, e_prio = r e in
	    sprintf "%s(%s)" (if e_prio < 998 then "("^e_^")" else e_) (r_ e'), 998
	| Some op -> 
	    match e' with
	    | Sep(",", l)
	    | SepL(",", _, l) -> handle_op op (map r l)
	    | _ -> handle_op op [r e'])
    | Float(f,_) -> string_of_float f, 999
    | String(s,_) -> sprintf "\"%s\"" s, 998
    | Sep(",", [e]) -> sprintf "((%s))" (r_ e), 998
    | Sep(sep,l)
    | SepL(sep,_,l) -> handle_op sep (map r l)
    | External(name,_) -> "External(" ^ name ^ ")", 998
    | Local ((i,stack,v),_) -> (if verbose || v = "" then sprintf "Local(%d,%d,%s)" i stack v else v), 999
    | Global ((i,pkg,v),_) -> 
	let v' = pkg ^ "::" ^ v in
	(if verbose || v = "" then sprintf "Global(%d,%s)" i v' else v'), 999
    | Func(_, para, body) -> handle_op "->" [ handle_op "," (map r para) ; r body ]
    | Ref(e,n,_) -> (if deref then sprintf "ref(%s is %s)" n (r_ !e) else sprintf "ref(%s)" n), 998
  in r_

let rew2s e = rew2s_ verbose false e

let constr2s = function
  | TFree -> "TFree"
  | TExact(Call(Constr("Pervasives::Builtin", "Open_function", _), t)) -> "Open " ^ rew2s t
  | TExact(t) -> sprintf "Exact(%s)" (rew2s t)
  | TConstr { lower = lo ; greater = gr } -> 
      let lo', gr' = map_t2 rew2s (lo, gr) in
      (if lo' = "Subval" then "" else sprintf "> (%s) " (rew2s lo)) ^
      (if gr' = "Object" then "" else sprintf "< (%s) " (rew2s gr))

let lattice2s latt =
  let elem2s (e, bounds) = Printf.sprintf "  %s (%s)\n" (rew2s e) (concat "  " (map (fun e -> "< " ^ rew2s e) bounds.Lattice.upper @ map (fun e -> "> " ^ rew2s e) bounds.Lattice.lower)) in
  concat "" (map elem2s latt)

let virtual_classes2s virtual_classes =
  let virtual_class2s (e, r) = Printf.sprintf "%s\n%s" (rew2s e) (lattice2s r.instanced_by) in
  let instanced_by = concat "" (map virtual_class2s virtual_classes.class2info) in
  let classes_lattice = lattice2s virtual_classes.lattice in
  instanced_by ^ "\n\n" ^ classes_lattice
