open Types
open List
open Common
open Printf

let unit = Constr("Pervasives::Builtin", "Empty_tuple", bpos)
let bottom = Constr("Pervasives::Builtin", "Bottom", bpos)
let anyval = Call(Constr("Pervasives::Builtin", "Var_forall", bpos), SepL(";", [ [] ], [ Local((0, 0, "_"), bpos) ]))
let superval = Call(Constr("Pervasives::Builtin", "All_T", bpos), Sep(",", []))
let subval = Call(Constr("Pervasives::Builtin", "Or", bpos), Sep(",", []))
let abstract pkg s = Call(Constr("Pervasives::Builtin", "Abstract", bpos), Constr(pkg, s, bpos))
let abstract_strict = abstract "Pervasives::Builtin" "Strict"

let debug = Flags.debug_rew

let reff x = ref (ref x)

let var_depth = function
  | Local((_,stack,_),_) -> Some stack
  | Global _ -> Some (99999)
  | _ -> None

let var_indice = function
  | Local((i,_,_),_) -> i
  | _ -> internal_error "var_indice"

let var_package = function
  | Global((_, pkg, _), _) -> pkg
  | _ -> internal_error "var_package"

let rec deref = function
  | Ref(r,_,_) -> deref !r
  | e -> e

let rec allt2l e = 
  match e with
  | Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", l)) -> collect allt2l l
  | e -> [e]
let rec or2l e = 
  match e with
  | Call(Constr("Pervasives::Builtin", "Or", _), Sep(",", l)) -> collect or2l l
  | e -> [e]

let l2allt l = 
  match collect allt2l l with
  | [e] -> e
  | l -> Call(Constr("Pervasives::Builtin", "All_T", bpos), Sep(",", l))
let l2or l = 
  match collect or2l l with
  | [e] -> e
  | l -> Call(Constr("Pervasives::Builtin", "Or", bpos), Sep(",", l))

let l2or_ l = if l = [] then [] else [ l2or l ]
let l2allt_ l = if l = [] then [] else [ l2allt l ]

let orallt2l = function
| Constr("Pervasives::Builtin", "Or", _) -> or2l
| Constr("Pervasives::Builtin", "All_T", _) -> allt2l
| _ -> failwith "orallt2l"

let l2orallt = function
| Constr("Pervasives::Builtin", "Or", _) -> l2or
| Constr("Pervasives::Builtin", "All_T", _) -> l2allt
| _ -> failwith "l2orallt"

let allt2l_func e = 
  collect (function
    | Func(nb, para, body) -> map (fun body -> Func(nb, para, body)) (allt2l body)
    | e -> [e]
  ) (allt2l e)

let allt_expand_call = function
  | Call(Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", l)), para) ->
      map (fun c -> Call(c, para)) l
  | e -> [e]

let rec is_same a b = 
  match deref a, deref b with
  | Float(a,_), Float(b,_) -> a = b
  | String(a,_), String(b,_) -> a = b
  | Constr(qa,ea,_), Constr(qb,eb,_) -> qa = qb && ea = eb
  | Local((ia,stacka,_),_), Local((ib,stackb,_),_) -> ia = ib && stacka = stackb
  | Global((pkga,ia,_),_), Global((pkgb,ib,_),_) -> ia = ib && pkga = pkgb
  | External(a,_), External(b,_) -> a = b

  | Func(_, pa, ba), Func(_, pb, bb) -> is_same_l (ba::pa) (bb::pb)
  | Func _, _
  | _, Func _ -> false (*failwith "is_same: Func can't be compared"*)

  | Call(fa,pa), Call(fb,pb) -> is_same fa fb && is_same pa pb

  | Sep(sa,la), Sep(sb,lb) -> sa = sb && is_same_l la lb
  | SepL(sa,_,la), SepL(sb,_,lb) -> sa = sb && is_same_l la lb

  | _, _ -> false

and is_same_l l = for_all2_ is_same l

let rec is_const = function
  | Float _    -> true
  | String _   -> true
  | Constr _ -> true
  | Call(f,e)     -> is_const f && is_const e
  | Sep(_, l)
  | SepL(_, _, l) -> for_all is_const l
  | _ -> false

let is_special_constr = function
  | Constr(_, "Inout", _)
  | Constr("Pervasives::Builtin", "All_T", _)
  | Constr("Pervasives::Builtin", "Or", _)
  | Constr("Pervasives::Builtin", "Or_range", _)
  | Constr("Pervasives::Builtin", "Abstract", _)
  | Constr("Pervasives::Builtin", "Val", _)
  | Constr("Pervasives::Builtin", "Var_forall", _)
  | Constr("Pervasives::Builtin", "Var_exists", _) -> true
  | _ -> false

let rec is_simple_const = function
  | Constr _ as c -> not (is_special_constr c)
  | Float _    -> true
  | String _   -> true
  | Call(f,e)     -> is_simple_const f && is_simple_const e
  | Sep(_, l)
  | SepL(_, _, l) -> for_all is_simple_const l
  | _ -> false

let is_func = function
  | Func _ -> true
  | _ -> false

let is_float = function
  | Float _ -> true
  | _ -> false

let is_string = function
  | String _ -> true
  | _ -> false

let is_char = function
  | String(s,_) -> String.length s = 1
  | _ -> false

let is_var = function
  | Local _ -> true
  | Global _ -> true
  | _ -> false

let is_local_var = function
  | Local _ -> true
  | _ -> false

let is_global_var = function
  | Global _ -> true
  | _ -> false

let is_external = function
  | External _ -> true
  | _ -> false

let is_capitalised_var = function
  | Local((_,_,n), _)
  | Global((_,_,n), _) -> n <> "" && n.[0] <> Char.lowercase n.[0]
  | _ -> false

let detype = function
  | Call(Constr("Pervasives::Builtin", "Type", _), e) -> e
  | e -> e

let is_allt = function
  | Constr("Pervasives::Builtin", "All_T", _) -> true
  | _ -> false

let is_or = function
  | Constr("Pervasives::Builtin", "Or", _) -> true
  | _ -> false

let is_orallt = function
  | Constr("Pervasives::Builtin", "All_T", _) -> true
  | Constr("Pervasives::Builtin", "Or", _) -> true
  | _ -> false

let is_abstract = function
  | Call(Constr("Pervasives::Builtin", "Abstract", _), _) -> true
  | _ -> false

let is_open_function = function
  | Call(Constr("Pervasives::Builtin", "Open_function", _), f) -> Some f
  | _ -> None

let is_strict = function 
  | Call(Constr("Pervasives::Builtin", "Abstract", _), Constr("Pervasives::Builtin", "Strict", _)) -> true
  | _ -> false

let is_dynamic = function
  | Call(Constr(_, "Dynamic", _), _) -> true
  | _ -> false

let is_superval = function
  | Call(Constr("Pervasives::Builtin", "All_T", _), Sep(",", [])) -> true
  | _ -> false

let is_subval = function
  | Call(Constr("Pervasives::Builtin", "Or", _), Sep(",", [])) -> true
  | _ -> false

let is_anyval = function
  | Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) -> true
  | _ -> false

let is_overload = function
  | Call(Constr("Pervasives::Builtin", "Overload", _), Sep(",", _)) -> true
  | _ -> false

let inout = function
  | Call(Constr(_, "Inout", _), _) as e -> e
  | e -> Call(Constr("_", "Inout", bpos), e)

let un_inout = function
  | Call(Constr(_, "Inout", _), e) -> e
  | e -> e

let l2or_option = function
  | [] -> None
  | l -> Some(l2or l)
let l2allt_option = function
  | [] -> None
  | l -> Some(l2allt l)

let or2or_option     e = if is_subval   e then None else Some e
let allt2allt_option e = if is_superval e then None else Some e

let overload l =
  let l' = map (fun (val_,typ) -> Sep(",", [ val_ ; typ ])) l in
  Call(Constr("Pervasives::Builtin", "Overload", bpos), Sep(",", l'))

let unoverload = function
  | e when is_same e bottom -> []
  | Call(Constr("Pervasives::Builtin", "Overload", _), Sep(",", l)) -> 
      map (function 
	| Sep(",", [ value ; typ ]) -> value, typ 
	| _ -> failwith "unoverload"
      ) l
  | _ -> failwith "unoverload"

let rec is_simple_const2 = function
  | Constr("Pervasives::Builtin", "Abstract", _) -> false
  | Constr("Pervasives::Builtin", "Type", _) -> false
  | Float _    -> true
  | String _   -> true
  | Constr _ -> true
  | Call(f,_)     -> is_simple_const2 f
  | Sep(_, l)
  | SepL(_, _, l) -> for_all is_simple_const2 l
  | Func(0, [any], body) when is_anyval any -> is_simple_const2 body
  | _ -> false

(* return the value *)
let unoverload_simple e =
  try 
    match unoverload e with
    | [ val_, _typ ] -> val_
    | _ -> failwith "unoverload_simple"
  with _ -> e

let unoverload_const e =
  try
    let vals = map fst (unoverload e) in
    if for_all is_simple_const vals then l2allt vals else e
  with _ -> e

let rec rew_map f e = 
  let e' = f e in
  match e' with
  | Call(e1, e2) -> Call(rew_map f e1, rew_map f e2)
  | Sep(s,l) -> Sep(s, map (rew_map f) l)
  | SepL(s,locals,l) -> SepL(s, locals, map (rew_map f) l)
  | Func(nb, para, body) -> Func(nb, map (rew_map f) para, rew_map f body)
  | e -> e

let rec rew_map_once f e = 
  match f e with
  | Some e' -> e'
  | None ->
      match e with
      | Call(e1, e2) -> Call(rew_map_once f e1, rew_map_once f e2)
      | Sep(s,l) -> Sep(s, map (rew_map_once f) l)
      | SepL(s,locals,l) -> SepL(s, locals, map (rew_map_once f) l)
      | Func(nb, para, body) -> Func(nb, map (rew_map_once f) para, rew_map_once f body)
      | e -> e

let rec rew_map_once_ f depth e = 
  match f(depth, e) with
  | Some e' -> e'
  | None ->
      match e with
      | Call(e1, e2) -> Call(rew_map_once_ f depth e1, rew_map_once_ f depth e2)
      | Sep(s,l) -> Sep(s, map (rew_map_once_ f depth) l)
      | SepL(s,locals,l) -> SepL(s, locals, map (rew_map_once_ f (depth+1)) l)
      | Func(nb, para, body) -> Func(nb, map (rew_map_once_ f (depth+1)) para, rew_map_once_ f depth body)
      | e -> e

let rec rew_iter_once f depth e = 
  if f depth e then () else
  match e with
  | Call(e1, e2) -> rew_iter_once f depth e1 ; rew_iter_once f depth e2
  | Sep(_,l) -> iter (rew_iter_once f depth) l
  | SepL(_,_,l) -> iter (rew_iter_once f (depth+1)) l
  | Func(_, para, body) -> iter (rew_iter_once f (depth+1)) para ; rew_iter_once f (depth+1) body
  | _ -> ()
	
let rew_map_ f e = 
  let rec rew_map_rec depth e =
    let e' = f(depth, e) in
    match e' with
    | Call(e1, e2) -> Call(rew_map_rec depth e1, rew_map_rec depth e2)
    | Sep(s,l) -> Sep(s, map (rew_map_rec depth) l)
    | SepL(s,locals,l) -> SepL(s, locals, map (rew_map_rec (depth+1)) l)
    | Func(nb, para, body) -> Func(nb, map (rew_map_rec (depth+1)) para, rew_map_rec (depth+1) body)
    | e -> e
  in
  rew_map_rec 0 e

let rew_map_withenv f env e = 
  let rec rew_map_rec depth env e =
    let e', env' = f env (depth, e) in
    match e' with
    | Call(e1, e2) -> 
	let (e1',e2'), env2 = map_t2_withenv (rew_map_rec depth) env' (e1, e2) in Call(e1', e2'), env2
    | Sep(s,l) -> 
	let l', env2 = map_withenv (rew_map_rec depth) env' l in Sep(s, l'), env2
    | SepL(s,locals,l) -> 
	let l', env2 = map_withenv (rew_map_rec (depth+1)) env' l in SepL(s, locals, l'), env2
    | Func(nb, para, body) -> 
	let para', env2 = map_withenv (rew_map_rec (depth+1)) env' para in
	let body', env3 = rew_map_rec (depth+1) env2 body in
	Func(nb, para', body'), env3
    | e -> e, env'
  in
  rew_map_rec 0 env e

let rec rew_map_shallow_withenv f env e = 
  let e', env' = f env e in
  match e' with
  | Call(e1, e2) -> 
      let (e1',e2'), env2 = map_t2_withenv (rew_map_shallow_withenv f) env' (e1, e2) in Call(e1', e2'), env2
  | Sep(s,l) -> 
      let l', env2 = map_withenv (rew_map_shallow_withenv f) env' l in Sep(s, l'), env2
  | e -> e, env'

let rec rew_map_leaf f e = 
  let rec rew_map_leaf_ depth f = function
    | Call(e1, e2) -> Call(rew_map_leaf_ depth f e1, rew_map_leaf_ depth f e2)
    | Sep(s,l) -> Sep(s, map (rew_map_leaf_ depth f) l)
    | SepL(s,locals,l) -> SepL(s, locals, map (rew_map_leaf_ (depth+1) f) l)
    | Func(nb, para, body) -> Func(nb, map (rew_map_leaf_ (depth+1) f) para, rew_map_leaf_ (depth+1) f body)
    | e -> f(depth,e)
  in
  rew_map_leaf_ 0 f e

let rec rew_map_leaf_valpart f e = 
  let rec rew_map_leaf_ depth f = function
    | Call(Constr("Pervasives::Builtin", "Overload", _), _) as e -> 
	failwith "unused336";
	overload (map (fun (v,t) -> rew_map_leaf_ depth f v, t) (unoverload e))
    | Call(e1, e2) -> Call(rew_map_leaf_ depth f e1, rew_map_leaf_ depth f e2)
    | Sep(s,l) -> Sep(s, map (rew_map_leaf_ depth f) l)
    | SepL(s,locals,l) -> SepL(s, locals, map (rew_map_leaf_ (depth+1) f) l)
    | Func(nb, para, body) -> Func(nb, map (rew_map_leaf_ (depth+1) f) para, rew_map_leaf_ (depth+1) f body)
    | e -> f(depth,e)
  in
  rew_map_leaf_ 0 f e

let rec rew_map_leaf_no_val f e = 
  let rec rew_map_leaf_no_val_ depth f = function
    | Call(Constr("Pervasives::Builtin", "Val", _), _) as e -> e
    | Call(e1, e2) -> Call(rew_map_leaf_no_val_ depth f e1, rew_map_leaf_no_val_ depth f e2)
    | Sep(s,l) -> Sep(s, map (rew_map_leaf_no_val_ depth f) l)
    | SepL(s,locals,l) -> SepL(s, locals, map (rew_map_leaf_no_val_ (depth+1) f) l)
    | Func(nb, para, body) -> Func(nb, map (rew_map_leaf_no_val_ (depth+1) f) para, rew_map_leaf_no_val_ (depth+1) f body)
    | e -> f(depth,e)
  in
  rew_map_leaf_no_val_ 0 f e

let rec rew_map_leaf__option f e = 
  let rec rew_map_leaf_ depth f = function
    | Call(e1, e2) -> map_option (fun (e1,e2) -> Call(e1,e2)) (t2_option2option_t2(rew_map_leaf_ depth f e1, rew_map_leaf_ depth f e2))
    | Sep(s,l) -> map_option (fun l' -> Sep(s, l')) (l_option2option_l (map (rew_map_leaf_ depth f) l))
    | SepL(s,locals,l) -> map_option (fun l' -> SepL(s, locals, l')) (l_option2option_l (map (rew_map_leaf_ (depth+1) f) l))
    | Func(nb, para, body) -> map_option (fun (para',body') -> Func(nb, para', body')) (t2_option2option_t2(l_option2option_l (map (rew_map_leaf_ (depth+1) f) para), rew_map_leaf_ (depth+1) f body))
    | e -> f(depth,e)
  in
  rew_map_leaf_ 0 f e

let rec rew_collect f e =
  (f e) @ match e with
  | Call(e1, e2) -> rew_collect f e1 @ rew_collect f e2
  | Sep(_,l) -> collect (rew_collect f) l
  | SepL(_,_,l) -> collect (rew_collect f) l
  | _ -> []

let rew_collect_ f e =
  let rec rew_collect_rec depth e =
    (f(depth,e)) @ match e with
    | Call(e1, e2) -> rew_collect_rec depth e1 @ rew_collect_rec depth e2
    | Sep(_,l) -> collect (rew_collect_rec depth) l
    | SepL(_,_,l) -> collect (rew_collect_rec (depth+1)) l
    | Func(_,para,body) -> collect (rew_collect_rec (depth+1)) (body :: para)
    | _ -> []
  in rew_collect_rec 0 e

let rec rew_collect_leaf f e = 
  let rec rew_collect depth f = function
    | Call(Constr("Pervasives::Builtin", "Abstract", _), _) as e -> f(depth, e)
    | Call(Constr("Pervasives::Builtin", "Overload", _), _) as e -> collect (fun (v,_t) -> rew_collect depth f v) (unoverload e)
    | Call(e1, e2) -> rew_collect depth f e1 @ rew_collect depth f e2
    | Sep(_,l) -> collect (rew_collect depth f) l
    | SepL(_,_,l) -> collect (rew_collect (depth+1) f) l
    | Func(_, para, body) -> collect (rew_collect (depth+1) f) (body :: para)
    | e -> f(depth,e)
  in
  rew_collect 0 f e


let rec rew_collect_leaf_no_val f e = 
  let rec rew_collect depth f = function
    | Call(Constr("Pervasives::Builtin", "Val", _), _) -> []
    | Call(Constr("Pervasives::Builtin", "Abstract", _), _) as e -> f(depth, e)
    | Call(Constr("Pervasives::Builtin", "Overload", _), _) as e -> failwith "unused401"; collect (fun (v,_t) -> rew_collect depth f v) (unoverload e)
    | Call(e1, e2) -> rew_collect depth f e1 @ rew_collect depth f e2
    | Sep(_,l) -> collect (rew_collect depth f) l
    | SepL(_,_,l) -> collect (rew_collect (depth+1) f) l
    | Func(_, para, body) -> collect (rew_collect (depth+1) f) (body :: para)
    | e -> f(depth,e)
  in
  rew_collect 0 f e

let rew_remap_locals_ depth nb e = 
  if nb = 0 then e else
  rew_map_leaf (function 
    | depth', Local((i,stack,n),pos) when stack >= depth + depth' -> 
	if stack+nb < depth + depth' then (
	  if debug then printf "rew_remap_locals: cleaning out scoped access\n";
	  Constr("Pervasives::Builtin", "Bottom", bpos)
	 ) else Local((i,stack+nb,n),pos)
    | _, e -> e) e
let rew_remap_locals nb e = rew_remap_locals_ 0 nb e

let rew_replace_localvar ind depth value e =
  rew_map_leaf (function
    | depth', Local((i,stack,_),_) when i = ind && stack + depth' = depth ->
	rew_remap_locals depth' value
    | _, e -> e) e

let rew_collect_vars stack rew =
  rew_collect_leaf (function
    | depth, (Local((i, stack', n), pos)) when stack = stack' - depth -> [ Local((i, stack, n), pos) ]
    | _ -> []
  ) rew
let rew_collect_local_vars rew = map var_indice (rew_collect_vars 0 rew)
let rew_collect_vars_l stack l = collect (rew_collect_vars stack) l

let rew_collect_all_accessible_vars rew =
  rew_collect_leaf (function
    | depth, (Local((i, stack, n), pos)) when stack >= depth -> [ Local((i, stack - depth, n), pos) ]
    | _, (Global _ as var) -> [ var ]
    | _ -> []
  ) rew

let rew_collect_all_accessible_local_vars rew =
  rew_collect_leaf (function
    | depth, (Local((i, stack, n), pos)) when stack >= depth -> [ Local((i, stack - depth, n), pos) ]
    | _ -> []
  ) rew

let rew_collect_all_accessible_vars_no_val rew =
  rew_collect_leaf_no_val (function
    | depth, (Local((i, stack, n), pos)) when stack >= depth -> [ Local((i, stack - depth, n), pos) ]
    | _, (Global _ as var) -> [ var ]
    | _ -> []
  ) rew

let rec getpos = function
  | Float(_,pos) -> pos
  | String(_,pos) -> pos
  | Constr(_,_,pos) -> pos
  | Local(_,pos) -> pos
  | Global(_,pos) -> pos
  | External(_,pos) -> pos
  | Ref(_,_,pos) -> pos
  | Func(_,para,body) -> uniposl (map getpos (body :: para))
  | Call(e1, e2) -> unipos (getpos e1) (getpos e2)
  | Sep(_,l) -> uniposl (map getpos l)
  | SepL(_,_,l) -> uniposl (map getpos l)

let sep2l = function
  | Sep(",", l) -> l
  | _ -> failwith "sep2l"

let sep2l_ = function
  | Constr("Pervasives::Builtin", "Empty_tuple", _) -> []
  | Sep(",", l) -> l
  | e -> [e]

let sep2non_empty_l = function
  | Sep(",", l) -> l
  | e -> [e]

let l2sep = function
  | [] -> Constr("Pervasives::Builtin", "Empty_tuple", bpos)
  | l -> Sep(",", l)
let l2sep_ = function
  | [] -> Constr("Pervasives::Builtin", "Empty_tuple", bpos)
  | [e] -> e
  | l -> Sep(",", l)

let sep2nb = function
  | Sep(",", l) -> length l
  | SepL(",", _, l) -> length l
  | _ -> 1

let unlazy_and_alike = function
  | Call(Constr(_, "Lazy", _), e) -> e, Some `Lazy
  | Call(Constr(_, "Dynamic", _), e) -> e, Some `Dynamic
  | e -> e, None

let lazy_and_alike2rew e = function
  | None -> e
  | Some flag -> 
      let s = match flag with
      |	`Lazy -> "Lazy"
      |	`Dynamic -> "Dynamic"
      in Call(Constr("_", s, bpos), e)

let rec is_abstract2 = function
  | Call(Constr("Pervasives::Builtin", "Abstract", _), _)
  | Call(Constr("Pervasives::Builtin", "Var_forall", _), _)
  | Call(Constr("Pervasives::Builtin", "Var_exists", _), _) -> true
  | Call(c, _) as e when is_allt c -> for_all is_abstract2 (allt2l e)
  | Call(c, _) -> is_abstract2 c
  | Func(0, [any], body) when is_anyval any -> is_abstract2 body
  | Func(1, [ Local((0,0,_),_) ], body) -> is_abstract2 body
  | Sep(";", l) -> is_abstract2 (last l)
  | _ -> false

let is_abstract2_bounded =
  let rec is_abstract2_bounded_ = function
    | Call(Constr("Pervasives::Builtin", "Var_forall", _), SepL(";", [_], [ Local((0, 0, "_"), _) ])) -> false
    | Call(Constr("Pervasives::Builtin", "Abstract", _), _)
    | Call(Constr("Pervasives::Builtin", "Var_forall", _), _)
    | Call(Constr("Pervasives::Builtin", "Var_exists", _), _) -> true
    | Call(c, _) as e when is_allt c -> for_all is_abstract2_bounded_ (allt2l e)
    | Func(0, [any], body) when is_anyval any -> is_abstract2_bounded_ body
    | Func(1, [ Local((0,0,_),_) ], body) -> is_abstract2_bounded_ body
    | _ -> false
  in function
    | Func(0, [any], _) when is_anyval any -> false
    | e -> is_abstract2_bounded_ e

let rec no_abstract = function
  | Call(Constr("Pervasives::Builtin", "Abstract", _), _)
  | Call(Constr("Pervasives::Builtin", "Var_forall", _), _)
  | Call(Constr("Pervasives::Builtin", "Var_exists", _), _) -> false
  | Call(c, e) -> no_abstract c && no_abstract e
  | Func(_, para, body) -> for_all no_abstract para && no_abstract body
  | Sep(_, l) -> for_all no_abstract l
  | _ -> true

let superval2anyval e = if is_superval e then anyval else e

let bool2rew b = Constr("Pervasives::Builtin", (if b then "True" else "False"), bpos)

let rew2bool = function
  | Constr("Pervasives::Builtin", "True", _) -> true
  | Constr("Pervasives::Builtin", "False", _) -> false
  | _ -> failwith "bad return value of condition (neither True nor False)"

let global2name = function
  | Global((_, pkg, n), _) -> pkg ^ "::" ^ n
  | _ -> failwith "global2name"

let rec rew_compare a b =
  match a,b with
  | Float(a,_),      Float(b,_)       -> compare a b
  | String(a,_),     String(b,_)      -> compare a b
  | External(a,_),   External(b,_)    -> compare a b
  | Global((pkga,a,_),_), Global((pkgb,b,_), _) -> 
      let r = compare pkga pkgb in if r = 0 then compare a b else r

  | Local((a2,a1,_), _), Local((b2,b1,_), _) -> compare_lists compare [ a1 ; a2 ] [ b1 ; b2 ]
  | Constr(a1,a2,_),     Constr(b1, b2, _)   -> compare_lists compare [ a1 ; a2 ] [ b1 ; b2 ]
  | Call(a1,a2),         Call(b1, b2)        -> compare_lists rew_compare [ a1 ; a2 ] [ b1 ; b2 ]

  | Func(_, pa, ba), Func(_, pb, bb) -> compare_lists rew_compare (ba :: pa) (bb :: pb)

  | Sep(_, la), Sep(_, lb)
  | SepL(_, _, la) , SepL(_, _, lb) -> compare_lists rew_compare la lb

  | Float _, _    -> -1 | _, Float _    -> 1
  | String _, _   -> -1 | _, String _   -> 1
  | Constr _, _   -> -1 | _, Constr _   -> 1
  | Local _, _    -> -1 | _, Local _    -> 1
  | Global _, _   -> -1 | _, Global _   -> 1
  | External _, _ -> -1 | _, External _ -> 1
  | Call _, _     -> -1 | _, Call _     -> 1
  | Func _, _     -> -1 | _, Func _     -> 1
  | Sep _, _      -> -1 | _, Sep _      -> 1
  | SepL _, _     -> -1 | _, SepL _     -> 1
  | Ref _, _      -> -1

let rew_sort l = sort rew_compare l
    

let rew_pack_locals l = 
  let vars = map (function Local((i,_,n),_) -> n,i | _ -> failwith "rew_pack_locals") (rew_collect_vars_l 0 l) in
  let best = function
    | (n,i), (n',i') when i = i' ->
	let n'' = if n = "" || n = "_" || String.length n > String.length n' then n' else n in
	Some ((if n'' = "" || n'' = "_" then "z_" else n''), i)
    | _ -> None
  in
  let vars2 = keep_best best vars in

  (* for different variable with same name, choose a different name *)
  let vars3 = collect (fun (n,i_s) ->
    fst (map_withenv (fun postfix i -> (n ^ postfix, i), postfix ^ "'") "" i_s)
  ) (prepare_want_all_assoc vars2) in

  let simplify_names l =
    let rec f accu = function
      | [] -> accu
      | (n,i)::l -> 
	  let rec simpl n =
	    let n' = chop n in
	    if n' = "" || mem n' (map fst (accu @ l)) then n 
	    else simpl n'
	  in f ((simpl n, i) :: accu) l
    in rev (f [] l) in
  let vars4 = fix_point simplify_names vars3 in

  let vars5 = map_index (fun (n,i) i' -> i, (n,i')) vars4 in
  let l' = map (rew_map_leaf (function 
    | depth, Local((i,stack,_),pos) when stack = depth -> 
	let (n',i') = assoc i vars5 in Local((i',stack,n'),pos)
    | _, e -> e)) l
  in l', length vars5


(* sort function according to the usage graph: if ``f'' uses ``g'', g will be after f *)
let rec sort_defs = 
  let rec find_it i lb = function
    | [] -> None
    | (Local((i',0,_),_), _, _) as e :: l when i=i'->
	Some(e, lb @ l)
    | e :: l -> find_it i (lb @ [e]) l
  in	
  function
  | [] -> []
  | (_,_,deps) as e :: l ->
      let (sorted, l') = fold_left (fun (sorted,l) dep ->
	match find_it dep [] l with
	| None -> sorted,l
	| Some(e, l') -> sorted @ [e], l'
      ) ([],l) deps
      in 
      sorted @ [e] @ sort_defs l'

let rec rew_remove_unused_sepl e = 
  let unseq = function Sep(";", [e]) -> e | e -> e in
  rew_map_once (function
    | Call(Constr("Pervasives::Builtin", "Var_forall", _), _)
    | Call(Constr("Pervasives::Builtin", "Var_exists", _), _) as e -> Some e (* don't modify it *)

    | SepL(s, [], l) -> Some (rew_remap_locals (-1) (unseq (Sep(s, map rew_remove_unused_sepl l))))

    | SepL(s, locals, l) ->
	(* trying to use "anyval"s *)
	let vars_used = collect rew_collect_local_vars l in
	if for_all (fun (_,x) -> x <= 1) (count_uniq vars_used) then
	  (* we can replace local vars with "anyval"s *)
	  let l' = map (rew_map_leaf (function 
	    | depth, Local((i,stack,n),pos) when stack >= depth -> 
		if stack = depth then anyval else Local((i,stack-1,n),pos)
	    | _, e -> e)) (map rew_remove_unused_sepl l) in
	  Some (unseq (Sep(s, l')))
	else if s = ";" && for_all ((=) 0) vars_used && (match last l with Local((0, 0, _), _) -> true | _ -> false) then
	  Some(Call(Constr("Pervasives::Builtin", "Var_forall", bpos), SepL(";", locals, map rew_remove_unused_sepl l)))
	else None

    | Func(nblocal, para, body) ->
	let nb = ref nblocal in
	(* flattening the variables in body *)
	let body' = 
	  rew_map_once (function
	    | Call(Constr("Pervasives::Builtin", "Var_forall", _), _)
	    | Call(Constr("Pervasives::Builtin", "Var_exists", _), _) as e -> Some e (* don't modify it *)
	    | SepL(s, locals, l) ->
		let l = if s = ";" then fix_point (function Constr("Pervasives::Builtin", "Empty_tuple", _) :: (_ :: _ as l) -> l | l -> l) l else l in
		let l' = map (rew_map_leaf (function 
		  | depth, Local((i,stack,n),pos) when depth = 0 || stack >= depth -> 
		      Local((if stack = depth then i + !nb, stack, n else i, stack-1, n), pos)
		  | _, e -> e)) (map rew_remove_unused_sepl l) in
		nb := !nb + length locals ;
		Some(Sep(s, l'))
	    | _ -> None) body in
	Some(Func(!nb, para, body'))
    | _ -> None) e

let rec may_be_a_func = function
  | External _
  | Func _
  | Call(Constr("Pervasives::Builtin", "Overload", _), _) -> true
  | _ -> false

let has_dynamic_para e = 
  let l = (try map snd (unoverload e) with _ -> [e]) in
  exists (function
    | Func(_, p, _) ->
	exists is_dynamic p
    | _ -> false
  ) l

let create_sequence l t =
  if l = [] then t else Sep(";", l @ [t])

let add_packed_constr l t = create_sequence l t

let l2orallt_factorize_constr funct l = 
  let constraints, l' = split (map (function
    | Sep(";", l) -> split_last l
    | e -> [], e
  ) l) in
  add_packed_constr (flatten constraints) (l2orallt funct l')

let compute_nbpara_bounds in_para =
  let nb = length in_para in
  match last in_para with
  | Call(Constr("Pervasives::Builtin", "Remaining", _), _) -> nb-1, `NoLimit
  | _ -> nb, `Strict

(* useful for adjusting the return_context when calling a function with
  variable number of arguments.
  replacing the Remaining(_) with as many "In" as necessary *)
let adjust_para para in_para =
  let nb = length in_para in
  let l, last = split_last in_para in
  let l' = match last with
  | Call(Constr("Pervasives::Builtin", "Remaining", _), _) -> l @ repeat (Constr("_", "In", bpos)) (length (sep2l_ para) - nb + 1)
  | _ -> in_para
  in Some (l2sep l')

let less_io = function
  | (io1, io2) when io1=io2 -> true
  | (In, Exact) -> true
  | (In, Inout) -> true
  | (Exact, Inout) -> true
  | (Out, Inout) -> true
  | _ -> false

let max_io io1 io2 = 
  match io1, io2 with
  | In, In -> In
  | Out, Out -> Out
  | Exact, Exact
  | Exact, In 
  | In, Exact -> Exact
  | _ -> Inout
let rew2io = function
  | Ref(r,_,_) -> Inout, !r
  | Call(Constr(_, "Out",     _), e) -> Out, e
  | Call(Constr(_, "Inout",   _), e) -> Inout, e
  | Call(Constr(_, "Exact",   _), e) -> Exact, e
  | e -> In, e

let io2rew e io =
  let (_io', e) = rew2io e in
  match io with
  | Out     -> Call(Constr("_", "Out", bpos), e)
  | Exact   -> Call(Constr("_", "Exact", bpos), e)
  | Inout   -> Call(Constr("_", "Inout", bpos), e)
  | In -> e

let inv_io = function
  | In -> Out
  | Out -> In
  | Inout -> Inout
  | Exact -> Exact

let rew_collect_with_io f (io, e) =
  let rec rew_collect_with_io_rec io depth e =
    match e with
    | Call(Constr(_, "Out",     _), e) -> rew_collect_with_io_rec (inv_io io) depth e
    | Call(Constr(_, "Inout",   _), e) -> rew_collect_with_io_rec Inout depth e
    | Call(Constr(_, "Exact",   _), e) -> rew_collect_with_io_rec (max_io Exact io) depth e
    | Call(e1, e2) -> rew_collect_with_io_rec io depth e1 @ rew_collect_with_io_rec io depth e2
    | Sep(_,l) -> collect (rew_collect_with_io_rec io depth) l
    | SepL(_,_,l) -> collect (rew_collect_with_io_rec io (depth+1)) l
    | Func(_,para,body) -> rew_collect_with_io_rec io (depth+1) body @ collect (rew_collect_with_io_rec (inv_io io) (depth+1)) para
    | _ -> f(io,depth,e)
  in rew_collect_with_io_rec io 0 e



let mark_as_val_simple e = Call(Constr("Pervasives::Builtin", "Val", bpos), e)

(* only mark non local *)
let rec mark_as_val e =
  let e' = rew_map_leaf_valpart (fun (depth, e) ->
    if match e with
    | Local((_,stack,_),_) when stack >= depth -> true
    | Global _ -> true
    | External _ -> true
    | _ -> false
    then mark_as_val_simple e else e
  ) e

  in rew_map (function
    | Call(Constr("Pervasives::Builtin", "Val", _), Call(Constr("Pervasives::Builtin", "Val", _), _)) -> failwith "mark_as_val"
    | Call(Constr("Pervasives::Builtin", "Type", _), Call(Constr("Pervasives::Builtin", "Val", _), e)) -> e
    |  e -> e
  ) e'

(* only mark non local *)
let rec mark_as_typ e =
  let e' = rew_map_leaf_valpart (fun (depth, e) ->
    if match e with
    | Local((_,stack,_),_) when stack >= depth -> true
    | Global _ -> true
    | External _ -> true
    | _ -> false
    then Call(Constr("Pervasives::Builtin", "Type", getpos e), e) else e
  ) e
	  
  in rew_map (function
    | Call(Constr("Pervasives::Builtin", "Type", _), Call(Constr("Pervasives::Builtin", "Type", _), _)) -> failwith "mark_as_typ"
    | Call(Constr("Pervasives::Builtin", "Val", _), Call(Constr("Pervasives::Builtin", "Type", _), e)) -> e
    |  e -> e
  ) e'

let constr_is_open_function = function
  | TExact(t) -> exists (fun e -> is_open_function e <> None) (allt2l t)
  | _ -> false

let rec full_deref = function
  | Call(Constr("Pervasives::Builtin", "Ref", _), e) -> e
  | Call(e1, e2) -> Call(full_deref e1, full_deref e2)
  | Sep(s,l) -> Sep(s, map full_deref l)
  | SepL(s,locals,l) -> SepL(s, locals, map full_deref l)
  | Func(nb, para, body) -> Func(nb, map full_deref para, full_deref body)
  | Ref(r,_,_) -> full_deref !r
  | e -> e

let rec full_deref_no_func = function
  | Call(Constr("Pervasives::Builtin", "Ref", _), e) -> e
  | Call(e1, e2) -> Call(full_deref_no_func e1, full_deref_no_func e2)
  | Sep(s,l) -> Sep(s, map full_deref_no_func l)
  | SepL(s,locals,l) -> SepL(s, locals, map full_deref_no_func l)
  | Ref(r,_,_) -> full_deref_no_func !r
  | e -> e

let flatten_tuple e =
  let flatten_tuple_ = function
    | Call(Constr("Pervasives::Builtin", "Flatten_tuple", _), Sep(",", l)) -> l
    | Call(Constr("Pervasives::Builtin", "Flatten_tuple", _), _) -> failwith "flatten_tuple"
    | e -> [e]
  in
  match e with
  | Sep(",", l) -> Sep(",", collect flatten_tuple_ l)
  | e -> l2sep_ (flatten_tuple_ e)

let l2tuple = function
  | [] -> Constr("Pervasives::Builtin", "Empty_tuple", bpos)
  | l -> Sep(",", l)
let rec tuple2l_ = function
  | Call(Constr("Pervasives::Builtin", "Overload", _), _) as e -> 
      Some (find_some tuple2l_ (map fst (unoverload e)))
  | Call(Constr("Pervasives::Builtin", "All_T", _), _) as e ->
      Some (find_some tuple2l_ (allt2l e))
  | Constr("Pervasives::Builtin", "Empty_tuple", _) -> Some []
  | Sep(",", l) -> Some l
  | _ -> None
let tuple2l e = 
  match tuple2l_ e with
  | Some e' -> e'
  | None -> failwith "tuple2l"

let check_no_mixed_var_in_or e = 
  match e with
  | Call(Constr("Pervasives::Builtin", "Or", _), Sep(",", l)) ->
      if exists is_var l then failwith (sprintf "check failed: mixed expression %s contains a variable" (Print.rew2s e))
  | _ -> ()

