open Types
open Common
open Printf
open Print
open List
open Info
open Rew

let debug = Flags.debug_rewrite


let 
  builtin_op__assign,
  builtin_op__has_type,
  builtin_op__subtype,
  builtin_op__supertype,
  builtin_op__ref = 0, 1, 2, 3, 4

let builtin_ops = [ ":=" ; "!!" ; "!<" ; "!>" ; "ref" ]



type all_vars = { 
    package_name : string ; 
    local : (string * int) list list ; 
    global : (string * int) list ; 
    (*                   name      pkg      indice *)
    packages_imported : (string * (string * int)) list ;
    (*          pkg       full_pkg  name    indice *)
    packages : (string * (string * (string * int) list)) list ;
  }

let push vars = { vars with local = [] :: vars.local }
let pop vars = { vars with local = tl vars.local }

let isFqid = function
  | `Fqid _ -> true
  | _ -> false

let rec getpos = function
  | `Id(_,pos) -> pos
  | `Fqid(_,_,pos) -> pos
  | `Constr(_,pos) -> pos
  | `Fqconstr(_,_,pos) -> pos
  | `String(_,pos) -> pos
  | `Float(_,pos) -> pos
  | `Call(e1, e2) -> unipos (getpos e1) (getpos e2)
  | `Sep(_, l) -> uniposl (map getpos l)

let check_too_complex_type is_type expr = 
  if is_type <> None then failwith (sprintf "complex runtime expressions are not allowed when specifying a type.\nYou should use a type function%s" (pos2sfull (getpos expr)))


let addlocal vars pos v = 
  let i = length (hd vars.local) in 
  Local((i, 0, v), pos), { vars with local = (hd vars.local @ [v, i]) :: (tl vars.local) }
let addlocal_at_depth depth vars pos v = 
  let i = length (nth vars.local depth) in 
  Local((i, depth, v), pos), { vars with local = getset_nth vars.local depth (fun subl -> subl @ [v, i]) }
let addglobal vars pos v = 
  let i = length vars.global in 
  Global((i, vars.package_name, v), pos), { vars with global = vars.global @ [v, i] }

let isHole = function
  | `Fqconstr("Pervasives::Builtin", "Hole", _) -> true
  | _ -> false
let isFunc = function
  | `Call(`Fqid("", "->", _), _)
  | `Sep(";", `Call(`Fqid("", "->", _), _) :: _) -> true
  | _ -> false
let sep2l = function
  | `Sep(",", l) -> l
  | e -> [e]
let seq2l = function
  | `Sep(";", l) -> l
  | e -> [e]

let rec search_only_local nb limit_depth local pos v = match local with
| [] -> raise Not_found
| e::l -> 
    try Local((assoc v e, nb, v), pos)
    with Not_found -> 
      if limit_depth = Some nb then raise Not_found else
      search_only_local (nb+1) limit_depth l pos v

let search_local vars pos v = 
                    try Local ((assoc v (hd vars.local), 0, v), pos), vars
  with Not_found -> addlocal vars pos v

let search_local_at_depth depth vars pos v = 
                    try search_only_local 0 (Some depth) vars.local pos v, vars
  with Not_found -> addlocal_at_depth depth vars pos v

let search_any add vars pos v = 
                    try search_only_local 0 None vars.local pos v, vars
  with Not_found -> try Global((assoc v vars.global, vars.package_name, v), pos), vars
  with Not_found -> try let pkg, i = assoc v vars.packages_imported in Global((i, pkg, v), pos), vars
  with Not_found -> add vars pos v

let search_any' vars pos v =
  search_any (fun _ -> failwith (sprintf "searching ``%s'' failed%s" v (pos2sfull pos))) vars pos v

let search_any_s vars v =
  fst(search_any (fun _ -> failwith (sprintf "searching ``%s'' failed" v)) vars bpos v)

let search_any_constr vars pos v constr =
  search_any (fun _ _ _ -> constr, vars) vars pos v

let search_fully_qualified f vars pos name pkg =
  let real_pkg, package = 
    try assoc pkg vars.packages 
    with Not_found -> 
      (* faking an empty Pervasives::Builtin for Pervasives::Ops *)
      if pkg = "Pervasives::Builtin" then pkg, []
      else internal_error ("search_fully_qualified (missing package " ^ pkg) in
  (try Global((assoc name package, real_pkg, name), pos)
  with Not_found -> f real_pkg)

let search_fully_qualified_or_fail vars pos name pkg =
  let err _ = failwith (sprintf "``%s::%s'' doesn't exist%s" pkg name (pos2sfull pos)) in
  search_fully_qualified err vars pos name pkg

let rec e_map f e = 
  let e_map_ = function
  | `Call(e, v) -> `Call(e_map f e, e_map f v)
  | `Sep(sep, l) -> `Sep(sep, map (e_map f) l)
  | e -> f e
  in e_map_ (f e)

let e_map_withenv f env e = do_withenv e_map f env e

let rec e_collect f = function
  | `Call(e, v) -> e_collect f e @ e_collect f v
  | `Sep(_, l) -> collect (e_collect f) l
  | e -> f e

let rec (closure_decl : 'a -> int option -> all_vars -> fromgrouper -> fromrewrite * all_vars) = fun typ is_type vars e -> 
  if debug then printf "closure_decl %s\n" (e2s e);
  let close = closure_decl typ is_type in
  match e with
  | `Id(v,pos) ->
      (match typ with
      |	`FuncPattern -> search_local vars pos v
      |	`Local -> 
	  (match is_type with
	  | None -> search_any addlocal vars pos v
	  | Some depth ->
	      let rel_depth = length vars.local - depth - 1 in
	      search_local_at_depth rel_depth vars pos v)
      |	`Global -> search_any addglobal vars pos v)
  | `Constr(c,pos) -> 
      (match typ with
      | `FuncPattern -> search_any_constr vars pos c (Constr(vars.package_name, c, pos))
      |	`Local -> search_local vars pos c
      |	`Global -> search_any addglobal vars pos c)

  (* mainly for operators *)
  | `Fqid("",v,pos) -> 
      (match typ with
      | `FuncPattern -> search_any' vars pos v
      |	`Local -> search_local vars pos v
      |	`Global -> search_any addglobal vars pos v)

  | `Fqid(pkg,v,pos) -> search_fully_qualified_or_fail vars pos v pkg, vars
  | `Fqconstr(pkg,c,pos) ->
      search_fully_qualified (fun real_pkg -> Constr(real_pkg,c,pos)) vars pos c pkg, vars

  | `Call(`Fqid("", "->", _), `Sep(",", [para ; body])) -> 
      (match is_type with
      |	None ->
	  let vars' = push vars in
	  let para', vars' = closure_decl `FuncPattern is_type vars' para in
	  let body', vars' = closure_decl `FuncPattern is_type vars' body in
	  Func(length (hd vars'.local), sep2non_empty_l para', body'), pop vars'
      |	Some _ ->
	  let para', vars = closure_decl `FuncPattern is_type vars para in
	  let body', vars = closure_decl `FuncPattern is_type vars body in
	  Func(0, sep2non_empty_l (rew_remap_locals 1 para'), rew_remap_locals 1 body'), vars)

    (* special case needed by "unclass" *)
  | `Call(`Fqconstr("Pervasives::Builtin", "Abstract", pos1), `Constr(c, pos)) ->
      Call(Constr("Pervasives::Builtin", "Abstract", pos1), Constr(vars.package_name, c, pos)), vars

  | `Call(e1, e2) ->
      map_fst (fun l -> Call(nth l 0, nth l 1)) (map_withenv close vars [e1 ; e2])
  | `Sep(e, l) -> 
      map_fst (fun l -> Sep(e, l)) (map_withenv close vars l)

  | `String(e,pos) -> String(e,pos), vars
  | `Float(e,pos) -> Float(e,pos), vars

let analyse_para is_type (vars, holes) = function
  | `Id(v,pos) -> 
      let (i, vars') = addlocal vars pos v in i, (vars', holes)
  | `Fqconstr("Pervasives::Builtin", "Hole", pos) ->
      let (i, vars') = addlocal vars pos "h" in i, (vars', i :: holes)
  | e -> 
      let (v, vars') = closure_decl `FuncPattern is_type vars e in v, (vars', holes)
let analyse_paras is_type vars var = map_withenv (analyse_para is_type) (vars, []) (sep2l var)

let rec build_cond vars c v val_ = 
  match c, v with
  | Local ((_,_,_),pos), _ -> 
      (`All, Sep(";", [ Call(search_any_s vars ":=", Sep(",", [c;v])) ; Constr("Pervasives::Builtin", "True", pos) ])), val_
  | _ -> 
      let (varl, both) = fpartition (function Local((i,0,_),_), Local((i',d,_),_) -> Some(i,(i',d)) | _ -> None) (combine (sep2l_ c) (sep2l_ v)) in

      let val2 = rew_map_leaf (function 
	| depth, (Local((i,depth',n),pos) as v) when depth = depth' -> 
	    (try 
	      let i',d = assoc i varl in
	      Local((i', d + depth', n), pos)
	    with Not_found -> v)
	| _, e -> e) val_ in
      let (c',v') = split both in
      (if c' = [] 
      then `All, Constr("Pervasives::Builtin", "True", bpos) 
      else `Mixed(map is_local_var (sep2l_ c)), Call(search_fully_qualified_or_fail vars bpos "unify" "Pervasives::Builtin", 
						     Sep(",", [ l2sep_ v' ; 
								Call(Constr("Pervasives::Builtin", "Lvalue", bpos), l2sep_ c') ]))), val2

let lookfor_deconstruct vars e = 
  let force e = function
    | None -> e, id
    | Some(e,f) -> e, f
  in
  let rec aux = function
  | `Sep(",", l) ->
      let l' = map aux l in
      if for_all ((=) None) l' then
	None
      else
	let es, fs = split (map2 force l l') in
	Some (`Sep(",", es), (function Sep(",", l) -> Sep(",", map2 apply fs l) | _ -> failwith "lookfor_deconstruct"))
  | `Call(`Fqid("", v, pos), `Sep(",", l)) ->	
      (try
	let f, _ = search_any (fun _ -> raise Not_found) vars pos ("deconstruct_" ^ v) in
	Some (`Call(`Constr("Some", bpos), `Sep(",", l)), (fun in_ -> Call(f, in_)))
      with Not_found -> None)
  |  _ -> None
  in force e (aux e)
	
let unclass = function
  | `Call(`Fqid("", op, _) as assign, `Sep(",", [ left ; (right:fromgrouper) ])) when op = "=" || op = ":=" ->
      let rec do_unclass = function
	| `Call(`Id("class", _), intf) ->
	    let name, fname = match left with
	    | `Constr(n,_) -> n, id
	    | `Call(`Constr(n,_), v) -> n, fun f -> `Call(f, v)
	    | _ -> failwith ""
	    in

	    let class_ = fun pos -> `Call(`Fqconstr("Pervasives::Builtin", "Abstract", pos), `Constr(name, getpos left)) in
	    let intf2 = e_map (function 
	      |	`Constr("O", pos) -> class_ pos
	      |	e -> e
            ) intf in
	    let constr = fun o -> `Call(`Fqid("", "!<", bpos), `Sep(",", [ `Id(o, bpos) ; `Constr(name, bpos) ])) in
	    let intf3, methods = e_map_withenv (fun methods e ->
	      match e with
	      |	`Call(`Fqid("", op, pos), `Sep(",", [ method_ ; t ])) 
		when (op = "=" || op = ":=")
		->
		  let right, recons = match t with
		  | `Call(`Fqid("", "->", _) as funct, `Sep(",", [ left ; right ])) -> right, fun r -> `Call(funct, `Sep(",", [ left ; r ]))
		  | e -> e, id
		  in
		  let vars = uniq (e_collect (function `Id("o",_) -> ["o"] | `Id("o2",_) -> ["o2"] | _ -> []) t) in
		  let right' = 
		    if vars = [] then right
		    else
		      `Sep(";", map constr vars @ seq2l(right)) in
		  `Call(`Fqid("", "!!", pos), `Sep(",", [ method_ ; `Call(`Fqconstr("Pervasives::Builtin", "Open_function", pos), recons right') ])), method_ :: methods
	      | _ -> e, methods
            ) [] intf2 in	    
	    let proto = `Call(`Fqconstr("Pervasives::Builtin", "Virtual_class", bpos), `Sep(",", `Constr(name, bpos) :: methods)) in
	    fname(class_ (getpos intf)), (seq2l intf3, [proto])
	| `Call(f, v) -> map_fst (fun v' -> `Call(f, v')) (do_unclass v)
	| `Sep(sep, l) ->
	    let (l', both) = split (map do_unclass l) in
	    let (intf, protos) = split both in
	    `Sep(sep, l'), (flatten intf, flatten protos)
	| e -> e, ([], [])
      in 
      let (right', (interface, protos)) = do_unclass right in
      `Call(assign, `Sep(",", [ left ; right' ])) :: interface @ protos
  | e -> [e]


let rec pattern_matching' in_ vars (constr, val_) =
  let (constr', fvars) = closure_decl `FuncPattern None vars constr in
  let (val2, vars') = closure false None fvars val_ in
  let (always_succeeds, cond), val3 = build_cond vars constr' in_ val2 in
  SepL(",", repeat [] (length (hd fvars.local)), [ cond ; val3 ]), always_succeeds, vars'

and pattern_matching in_ (always_succeeds, vars) (constr, val_) =
  let constr', f = lookfor_deconstruct vars constr in
  let cond, always_succeeds', vars' = pattern_matching' (f in_) vars (constr', val_) in
  let always_succeeds'' = 
    match always_succeeds, always_succeeds' with
    | `All, _ -> eprintf "%s: Warning: match case unused\n" (pos2sfull' (getpos constr)) ; `All
    | `Mixed _, `All -> `All
    | `Mixed a, `Mixed b -> `Mixed (map2 (||) a b) in
  cond, (always_succeeds'', vars')

and pattern_matchings in1 vars patterns =
  let in_ = rew_remap_locals 1 (l2sep_ in1) in
  let conds, (always_succeeds, vars') = map_withenv (pattern_matching in_) (`Mixed(map (fun _ -> false) in1), push vars) patterns in
  let restricted_vars = 
    match always_succeeds with
    | `All -> []
    | `Mixed l -> snd (filter2 (fun (b, _) -> not b) l in1)
  in
  let packed_conds = Call(search_fully_qualified_or_fail vars bpos "cond" "Pervasives::Builtin", Sep(",", conds)) in
  let packed_conds' = 
    if restricted_vars = [] then packed_conds else 
    Call(Constr("Pervasives::Builtin", "Pattern_matching_vars", bpos), Sep(",", [ l2sep restricted_vars ; packed_conds ])) in
  packed_conds', pop vars'

and find_out_pattern = function
  | `Call(`Fqid("", "->", _), `Sep(",", [v ; va])) -> v, va
  | e -> failwith (sprintf "i should have a pattern matching, whereas i have %s%s" (e2s e) (pos2sfull (getpos e)))
and find_out_patterns = function
  | `Sep(";", l) -> map find_out_pattern l
  | e -> [ find_out_pattern e ]
and nb_para_of_pattern p = length (sep2l p)
and nb_para_of_patterns l = fold_left max 0 (map nb_para_of_pattern l)

and pre_declare_func typ is_type vars = function
  | `Call(`Fqid("", op, _), `Sep(",", [ left ; _right ])) when op = "=" || op = ":=" ->
      let var =
	match left with
	| `Call(`Id _ as var, _) -> Some var
	| `Id _ -> Some left
	| `Fqid _ when typ = `Global -> Some left
	| _ -> None
      in
      if var = None then vars else (
        check_too_complex_type is_type left ;
	if debug then printf "pre_declare_func %s\n" (e2s (some var));
	snd (closure_decl typ None vars (some var))
      )
  | _ -> vars

and (closure : bool -> int option -> all_vars -> fromgrouper -> fromrewrite * all_vars) = fun is_toplevel is_type vars expr -> 
  if debug then printf "closure %s\n" (e2s expr);
  match expr with
  | `Constr(e,pos) -> search_any_constr vars pos e (Constr(vars.package_name, e, pos))
  | `Fqconstr(pkg,c,pos) -> search_fully_qualified (fun real_pkg -> Constr(real_pkg,c,pos)) vars pos c pkg, vars
  | `Fqid("",v,pos) -> search_any' vars pos v
  | `Fqid(pkg,v,pos) -> search_fully_qualified_or_fail vars pos v pkg, vars

  | `String(e,pos) -> String(e,pos), vars
  | `Float(e,pos) -> Float(e,pos), vars

  | `Id("_",pos) -> addlocal vars pos "_"
  | `Id(e,pos) -> search_any' vars pos e


    (* special case needed by "unclass" *)
  | `Call(`Fqconstr("Pervasives::Builtin", "Abstract", pos1), `Constr(c, pos)) ->
      Call(Constr("Pervasives::Builtin", "Abstract", pos1), Constr(vars.package_name, c, pos)), vars

  | `Call(`Constr("External", pos), `String(e, pos2)) -> External(e, unipos pos pos2), vars

    (* non-anonymous function declaration *)
  | `Call(`Fqid("", op, pos), `Sep(",", [`Call(id, p) ; va])) 
    when (op = "=" || op = ":=") && not (isFqid id && not is_toplevel)
      ->
      check_too_complex_type is_type expr ;
      let v', vars' = closure_decl (if is_toplevel then `Global else `Local) None vars id in
      let (p', (vars'', holes)) = analyse_paras is_type (push vars') p in
      let (va', vars''') =
	if holes = [] then
	  closure false is_type vars'' va
	else
	  let patterns = find_out_patterns va in
	  let nbp = nb_para_of_patterns (map fst patterns) in
	  check_too_complex_type is_type expr ;
	  if (nbp <> length holes) then failwith (sprintf "bad nb holes (%d != %d)%s" nbp (length holes) (pos2sfull pos))
	  else
	    pattern_matchings holes vars'' patterns in

      (* HACK: function definition uses := instead of "=", easier for overloading and recursive functions *)
      Call(search_any_s vars ":=", Sep(",", [ v' ; Func(length (hd vars'''.local), p', va') ])), 
      pop vars'''

    (* variable affection (or function declaration via anonymous function) *)
  | `Call(`Fqid("", op, _) as affect, `Sep(",", [ left ; right ])) when op = "=" || op = ":=" -> 
      check_too_complex_type is_type expr ;
      let affect' =
	if isFunc right then (* HACK: function definition uses := instead of "=", easier for overloading and recursive functions *)
	  search_any_s vars ":="
	else 
	  fst(closure false is_type vars affect) in
      let (left', f_right) = lookfor_deconstruct vars left in
      let (left'', vars_) = closure_decl (if is_toplevel then `Global else `Local) None vars left' in
      let (right', vars') = closure false is_type vars_ right in
      Call(affect', Sep(",", [ left'' ; f_right right' ])), vars'

  | `Call(`Fqid("", t, _) as typ, `Sep(",", [ left ; right ])) when mem t [ "!!" ; "!>" ; "!<" ] ->
      let typ' = fst(closure false is_type vars typ) in
      let left', vars2 = 
	(* do not declare Point in Point(Double, Double) !< XXX *)
	(* do declare y in _ -> (y !> A ; y) *)
	let typ = 
	  if is_toplevel then 
	    match left with 
	    | `Fqid _
	    | `Id _ -> `Global
	    | _ -> `FuncPattern
	  else `Local in
	let left', vars = closure_decl typ is_type (push vars) left in
	let nblocals = length (hd vars.local) in
	(if nblocals = 0 then
	  rew_remap_locals (-1) left'
	else
	  SepL(";", repeat [] nblocals, [ left' ])), pop vars in
      let vars3 = push vars2 in
      let is_type' = if is_type = None then Some (length vars3.local) else is_type in
      let right', vars4 = closure false is_type' vars3 right in
      let right'' = SepL(";", repeat [] (length (hd vars4.local)), [ right' ]) in
      Call(typ', Sep(",", [ left' ; right'' ])), pop vars4


    (* anonymous function body *)
  | `Call(`Fqid("", "->", _), `Sep(",", [ left ; right ])) ->
      let vars' = push vars in
      let left', f_right = lookfor_deconstruct vars left in
      if left = left' then
	let (left'', vars_) = closure_decl `FuncPattern is_type vars' left' in
	let (right', vars'') = closure false is_type vars_ right in
	Func(length (hd vars''.local), sep2non_empty_l left'', right'), pop vars''
      else
	let vars'' = push vars in
	let nb = length (sep2l left') in
	let locals depth = map (fun i -> Local((i, depth, "h" ^ (if nb > 1 then string_of_int (i+1) else "")),bpos)) (range 0 nb) in
	check_too_complex_type is_type expr ;
	let (left'', vars_) = closure_decl `FuncPattern None vars'' left' in
	let (right', vars''') = closure false is_type vars_ right in
	let deconstruct = Call(search_any_s vars "=", Sep(",", [ left'' ; f_right (Sep(",", locals 1)) ])) in
	let func = Func(nb, locals nb, 
			SepL(";", repeat [] (length (hd vars''.local)), [ deconstruct ; right' ])) in
	func, pop vars'''
	     
    (* anonymous function with pattern matching *)
  | _ when isFunc expr ->
      check_too_complex_type is_type expr ;
      let patterns = find_out_patterns expr in
      let nbp = nb_para_of_patterns (map fst patterns) in
      let var_names = map (fun i -> "h" ^ (if nbp > 1 then string_of_int (i+1) else "")) (range 0 nbp) in
      let (paras, vars') = map_withenv (fun vars e -> addlocal vars bpos e) (push vars) var_names in
      let (val_, vars'') = pattern_matchings paras vars' patterns in
      Func(nbp, paras, val_), pop vars''

  | `Call(e, para) ->
      if exists isHole (sep2l para) then
	let (e', vars_) = closure is_toplevel is_type (push vars) e in
	let (nb, par) = ref 0, ref [] in
	match closure is_toplevel is_type vars_ para with
        | SepL(",", nbp, para'), vars' ->
	    let para'' = map (function
	      | Constr("Pervasives::Builtin", "Hole", pos) -> 
		  let p = Local((!nb, 1, "h"), pos) in
		  par := !par @ [ Local((!nb, 0, "h"), pos) ] ; 
		  incr nb ;
		  p
	      | e -> e) para' in
	    Func(!nb, !par, Call(e', SepL(",", nbp, para''))), pop vars'
	| Constr("Pervasives::Builtin", "Hole", pos), vars' -> 
	    let local = Local((0, 0, "h"), pos) in
	    (* eta expanded on purpose: partial application is NOT different from eta expansion *)
	    Func(1, [ local ], Call(e', local)), pop vars'
	| _ -> internal_error "Rewrite.closure"
      else
	let (e', vars_) = closure is_toplevel is_type vars e in
	let (para', vars') = closure is_toplevel is_type vars_ para in
	Call(e', para'), vars'

  | `Sep(sep, l) ->
      let l_unclass = collect unclass l in
      let vars' = fold_left (pre_declare_func (if is_toplevel then `Global else `Local) is_type) (push vars) l_unclass in 
      let (l', vars'') = map_withenv (closure is_toplevel is_type) vars' l_unclass in
      let l'' = 
	if sep = ";" && isHole(last l)
	then removelast l'
	else l' in
      SepL(sep, repeat [] (length (hd vars''.local)), l''), pop vars''


let rewrite package_name e imports packages_used all_packages = 
  let extract_from_p_globals pkg =
    Array.to_list (Array.mapi (fun i (_,_,_,n) -> !n, i) pkg.p_global) 
  in
  let packages =
    map (fun pkg -> pkg, (pkg, [])) [ "Pragma" ] @
    map_assoc_val (fun pkg -> pkg.full_name, extract_from_p_globals pkg) packages_used in

  let all_imports = collect (fun pkg -> 
    let l = map (fun s -> assoc_or_fail s all_packages) pkg.re_exported_packages in
    pkg :: l
  ) imports in
  let packages_imported = collect (fun pkg ->
      map (fun (name, i) -> name, (pkg.full_name, i)) (extract_from_p_globals pkg)
  ) all_imports in

  let globals =
    if package_name = "Pervasives::Builtin" then
      (* adding some things that can't be done any other way *)
      map_index (fun s i -> s, i) builtin_ops
    else [] in

  let vars = { local = [] ; global = globals ; package_name = package_name ; packages_imported = packages_imported ; packages = packages } in
  let (e', vars') = closure true None vars (`Sep(",", [e])) in
  e', map fst vars'.global
